'''
pip install opencv-python numpy
'''
import numpy as np
import cv2
import os
import datetime
import time # 用于计算处理时间

def format_timestamp(milliseconds):
    """将毫秒时间戳格式化为 H_M_S_ms 字符串"""
    if milliseconds < 0:
        milliseconds = 0
    td = datetime.timedelta(milliseconds=milliseconds)
    total_seconds = int(td.total_seconds())
    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    milliseconds_part = int(milliseconds % 1000)
    return f"{hours:02d}_{minutes:02d}_{seconds:02d}_{milliseconds_part:03d}"

def extract_subtitles(video_path, output_folder, roi_coords, threshold_value, change_threshold_percent):
    """
    从视频文件中提取指定区域内变化的硬字幕。

    Args:
        video_path (str): 输入视频文件路径 (MKV/HEVC)。
        output_folder (str): 保存提取出的字幕 PNG 图片的文件夹。
        roi_coords (tuple): 字幕区域坐标 (x, y, width, height)。
        threshold_value (int): 二值化阈值 (0-255)。用于区分背景和文字。
        change_threshold_percent (float): 变化百分比阈值 (0-100)。
                                          当二值化 ROI 与上一个保存的 ROI 相比，
                                          不同像素的百分比超过此值时，才保存新图片。
                                          例如，1.0 表示超过 1% 的像素变化就保存。
    """
    x, y, w, h = roi_coords

    # --- 安全性检查 ---
    if not os.path.exists(video_path):
        print(f"错误：视频文件未找到 - {video_path}")
        return
    if w <= 0 or h <= 0:
        print(f"错误：ROI 尺寸无效 (width={w}, height={h})")
        return
    if not (0 <= threshold_value <= 255):
         print(f"警告：二值化阈值 {threshold_value} 超出 [0, 255] 范围。")
    if not (0 < change_threshold_percent <= 100):
         print(f"警告：变化百分比阈值 {change_threshold_percent} 超出 (0, 100] 范围。推荐使用较小值如 1.0 或 5.0。")


    # --- 初始化 ---
    print("开始处理视频...")
    start_time = time.time()

    os.makedirs(output_folder, exist_ok=True)
    print(f"输出文件夹：{output_folder}")

    cap = cv2.VideoCapture(video_path)
    if not cap.isOpened():
        print(f"错误：无法打开视频文件 - {video_path}")
        print("请检查文件路径、文件是否损坏、以及 OpenCV 是否正确安装并支持 HEVC (通常是支持的)。")
        return

    # 获取视频基本信息 (可选，用于调试或信息展示)
    frame_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    fps = cap.get(cv2.CAP_PROP_FPS)
    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    print(f"视频信息: {frame_width}x{frame_height}, {fps:.2f} FPS, 总帧数: {total_frames if total_frames > 0 else '未知'}")

    # 再次检查ROI坐标是否在视频范围内
    if x < 0 or y < 0 or x + w > frame_width or y + h > frame_height:
        print(f"错误：ROI 坐标 {roi_coords} 超出视频边界 ({frame_width}x{frame_height})")
        cap.release()
        return

    last_saved_roi_binarized = None
    saved_image_count = 0
    processed_frame_count = 0
    roi_pixel_count = w * h

    # --- 逐帧处理 ---
    while True:
        ret, frame = cap.read()
        if not ret:
            print("视频处理完成。")
            break # 视频结束或读取错误

        processed_frame_count += 1
        timestamp_ms = cap.get(cv2.CAP_PROP_POS_MSEC)

        # 1. 提取 ROI
        roi = frame[y:y+h, x:x+w]

        # 2. 灰度化
        roi_gray = cv2.cvtColor(roi, cv2.COLOR_BGR2GRAY)

        # 3. 二值化 (白底黑字 -> 黑底白字，更适合差异比较)
        # THRESH_BINARY_INV: 低于阈值的为白色(255)，高于阈值的为黑色(0)
        # 如果你的原始字幕是亮色文字在暗背景上，可能需要 THRESH_BINARY
        # 根据你描述“白底黑字”但“视频其他地方很暗”，推测字幕区域本身可能是亮的
        # 如果目标是让文字变白，背景变黑，用 THRESH_BINARY 并调整阈值
        # 这里假设 threshold_value 是用来区分亮色文字和暗色背景的界限
        ret, roi_binarized = cv2.threshold(roi_gray, threshold_value, 255, cv2.THRESH_BINARY)
        # 如果你需要的是黑底白字，可以反转：
        # roi_binarized = cv2.bitwise_not(roi_binarized)

        # 4. 与上一个保存的 ROI 进行比较
        save_this_frame = False
        if last_saved_roi_binarized is None:
            # 第一个有效帧总是保存（如果它不是完全空白的话）
            # 可以加一个检查，比如非全黑或全白
            if np.count_nonzero(roi_binarized) > 0 and np.count_nonzero(roi_binarized) < roi_pixel_count: # 避免保存纯黑/纯白帧
                save_this_frame = True
        else:
            # 计算差异
            diff = cv2.absdiff(roi_binarized, last_saved_roi_binarized)
            non_zero_count = np.count_nonzero(diff)
            change_percentage = (non_zero_count / roi_pixel_count) * 100

            # 判断变化是否超过阈值
            if change_percentage > change_threshold_percent:
                save_this_frame = True

        # 5. 如果需要保存
        if save_this_frame:
            timestamp_str = format_timestamp(timestamp_ms)
            filename = os.path.join(output_folder, f"{timestamp_str}.png")

            # 检查是否与上一张保存的图片完全相同 (避免因微小时间差导致重复)
            # 这一步是可选的，但有时有帮助
            is_identical_to_last = False
            if last_saved_roi_binarized is not None:
                 if np.array_equal(roi_binarized, last_saved_roi_binarized):
                     is_identical_to_last = True

            if not is_identical_to_last:
                cv2.imwrite(filename, roi_binarized)
                # print(f"Saved: {filename} (Change: {change_percentage:.2f}%)") # 打印详细信息
                last_saved_roi_binarized = roi_binarized.copy() # 更新最后保存的图像 (必须用copy)
                saved_image_count += 1

        # (可选) 打印进度
        if processed_frame_count % 100 == 0: # 每处理100帧打印一次
            elapsed_time = time.time() - start_time
            print(f"已处理 {processed_frame_count} 帧, 已保存 {saved_image_count} 张字幕图片, 用时: {elapsed_time:.2f} 秒")


    # --- 清理 ---
    cap.release()
    # cv2.destroyAllWindows() # 如果中途开了窗口显示，需要这句

    end_time = time.time()
    print(f"\n处理完毕！总共处理了 {processed_frame_count} 帧。")
    print(f"共保存了 {saved_image_count} 张字幕图片到 {output_folder}")
    print(f"总耗时: {end_time - start_time:.2f} 秒")

# ==============================================================================
# ======================= 请修改以下参数 =======================================
# ==============================================================================

# 1. 输入视频文件路径
video_file = "movie.mkv" # <--- 修改这里

# 2. 输出文件夹路径
output_dir = "extracted_subtitles" # <--- 修改这里 (文件夹会自动创建)

# 3. 字幕区域坐标 (x, y, width, height) - 需要精确测量！
#    (x, y) 是矩形左上角的坐标，(0, 0) 在视频左上角
#    width 是矩形的宽度，height 是矩形的高度
roi_definition = (10, 1845, 3800, 120) # <--- !!!【极其重要】请根据你的视频精确测量并修改这里 (示例值) !!!

# 4. 二值化阈值 (0-255) - 需要实验调整！
#    目标是区分字幕文字和背景。如果字幕是纯白色，背景纯黑色，阈值可以在中间，比如 127。
#    如果字幕是白色，背景是灰色，可能需要更高的阈值，比如 200 或 220。
#    建议先用播放器截图，用图像编辑软件查看字幕和背景的像素值范围来辅助确定。
threshold = 240 # <--- !!!【重要】请根据你的视频实验并修改这里 (示例值) !!!

# 5. 变化百分比阈值 (0-100) - 需要实验调整！
#    当新字幕与上一个保存的字幕相比，像素差异超过这个百分比时，才保存。
#    设置太低可能会因噪点保存过多图片，设置太高可能会错过某些字幕更新。
#    1.0 或 5.0 是比较常用的起始值。99% 几乎意味着完全不同的图片，可能不是你想要的。
change_threshold = 1.5 # <--- !!!【重要】请根据你的视频实验并修改这里 (示例值：1.5%) !!!

# ==============================================================================
# ======================= 参数修改结束 =========================================
# ==============================================================================

# --- 运行提取函数 ---
if __name__ == "__main__":
    if not os.path.exists(video_file):
         print(f"错误：找不到视频文件 '{video_file}'。请确保路径正确。")
    else:
        extract_subtitles(
            video_path=video_file,
            output_folder=output_dir,
            roi_coords=roi_definition,
            threshold_value=threshold,
            change_threshold_percent=change_threshold
        )
        